import { strictEqual } from './equality';
import useEntity from './useEntity';
import alias from '../plugins/alias';
import { onInit, onSet } from '../tools/inspector';
import { getStore, isStoreEnabled } from '../tools/store';
function entity(initialValue, pluginsOrAlias) {
  if (pluginsOrAlias === void 0) {
    pluginsOrAlias = [];
  }
  if (initialValue === undefined) throw new Error('Entity requires an initial value.');
  var plugins = typeof pluginsOrAlias === 'string' ? [alias(pluginsOrAlias)] : pluginsOrAlias;
  if (!(plugins instanceof Array)) throw new Error('Invalid plug-ins array.');
  var entityBuild = {
    _value: undefined,
    _subscribers: new Set()
  };
  entityBuild.subscribe = createSubscribe(entityBuild);
  entityBuild.get = createGet(entityBuild);
  entityBuild.set = createSet(entityBuild);
  entityBuild.init = createInit(entityBuild, initialValue);
  entityBuild.use = createHook(entityBuild);
  var newEntity = entityBuild;
  applyPlugins(newEntity, plugins);
  newEntity.init();
  if (isStoreEnabled()) getStore().add(newEntity);
  return newEntity;
}
export default entity;
function createSubscribe(entity) {
  return function (subscriberFn) {
    entity._subscribers.add(subscriberFn);
    return function () {
      entity._subscribers.delete(subscriberFn);
    };
  };
}
function createGet(entity) {
  return function () {
    return entity._value;
  };
}
function createSet(entity) {
  return function (valueOrUpdaterFn, alias) {
    if (typeof valueOrUpdaterFn === 'function') valueOrUpdaterFn = valueOrUpdaterFn(entity._value);
    entity._value = valueOrUpdaterFn;
    entity._subscribers.forEach(function (cb) {
      return cb(entity._value);
    });
    if (alias !== '@@DEVTOOLS') onSet(entity, alias != null ? alias : '<anonymous>');
  };
}
var nextId = 1;
function createInit(entity, initialValue) {
  return function () {
    if (!entity.name) entity.name = "entity" + nextId++;
    if (initialValue instanceof Promise) initialValue.then(function (value) {
        return setTimeout(function () {
          return entity.set(value, '@@ASYNC_INIT');
        });
      });else entity._value = initialValue;
    onInit(entity);
  };
}
function createHook(entity) {
  return function (transform, equality) {
    if (transform === void 0) {
      transform = function transform(v) {
        return v;
      };
    }
    if (equality === void 0) {
      equality = strictEqual;
    }
    return useEntity(entity, transform, equality);
  };
}
function applyPlugins(entity, plugins) {
  plugins.forEach(function (plugin) {
    if (typeof plugin !== 'object') throw new Error('Invalid plug-in');
    function overrideMethod(method) {
      var createOverride = plugin[method];
      if (typeof createOverride === 'function') {
        var override = createOverride(entity[method], entity);
        if (typeof override !== 'function') throw new Error("Invalid override for '" + method + "' in plug-in.");
        entity[method] = override;
      }
    }
    Object.keys(plugin).forEach(function (method) {
      return overrideMethod(method);
    });
  });
}